﻿using System.Security.Cryptography;


namespace CNG
{
    /// <summary>
    ///     The CngProvider2 class provides additional <see cref="CngProvider" /> objects to suppliment the
    ///     ones found on the standard <see cref="CngProvider" /> type.
    /// </summary>
    internal static class CngProvider2
    {
        private static CngProvider s_primitiveAlgorithmProvider;

        /// <summary>
        ///     Get a CngProvider for the Microsoft Primitive algorithm provider
        /// </summary>
        public static CngProvider MicrosoftPrimitiveAlgorithmProvider
        {
            get
            {
                if (s_primitiveAlgorithmProvider == null)
                {
                    s_primitiveAlgorithmProvider = new CngProvider(BCryptNative.ProviderName.MicrosoftPrimitiveProvider);
                }

                return s_primitiveAlgorithmProvider;
            }
        }
    }

    /// <summary>
    ///     Well known BCrypt provider names
    /// </summary>
    internal static class ProviderName
    {
        internal const string MicrosoftPrimitiveProvider = "Microsoft Primitive Provider";      // MS_PRIMITIVE_PROVIDER
    }
}
